with text_io;                            use text_io;
with Standard_Floating_Numbers;          use Standard_Floating_Numbers;
with Standard_Floating_Vectors;
with Standard_Complex_Numbers;           use Standard_Complex_Numbers;
with Standard_Complex_Vectors;           use Standard_Complex_Vectors;
with Standard_Complex_Matrices;          use Standard_Complex_Matrices;
with Standard_Complex_Solutions;         use Standard_Complex_Solutions;

package Increment_Power_Continuation is

-- DESCRIPTION :
--   This package implements routines for tracing solution paths of a
--   polynomial homotopy that is nonlinear in the continuation parameter.
--   In particular, the following assumptions are made:
--     1) paths do not turn back, i.e.: increment-and-fix continuation;
--     2) the nonlinearity is due to high powers of the continuation parameter.
--   The reporting version writes intermediate results on file.

  generic

    with function Norm ( x : Vector ) return double_float;
    with function H  ( x : Vector; t : Complex_Number ) return Vector;
    with function dH ( x : Vector; t : Complex_Number ) return Matrix;

  procedure Silent_Power_Continue
               ( sols : in out Solution_List;
                 powers : in Standard_Floating_Vectors.Vector );

  generic

    with function Norm ( x : Vector ) return double_float;
    with function H  ( x : Vector; t : Complex_Number ) return Vector;
    with function dH ( x : Vector; t : Complex_Number ) return Matrix;

  procedure Reporting_Power_Continue
               ( file : in file_type; sols : in out Solution_List;
                 powers : in Standard_Floating_Vectors.Vector );

  -- DESCRIPTION :
  --   Increment-and-fix continuation with power predictor.

  -- ON ENTRY :
  --   file      to write intermediate results on (if Reporting_);
  --   sols      the start solutions;
  --   powers    vector of powers sorted in ascending order.

  -- ON RETURN :
  --   sols      the computed solutions.

end Increment_Power_Continuation;
