with text_io;                            use text_io;
with Standard_Complex_Numbers;           use Standard_Complex_Numbers;
with Standard_Complex_Solutions;         use Standard_Complex_Solutions;
with Standard_Complex_Poly_Systems;      use Standard_Complex_Poly_Systems;

package Drivers_for_Homotopy_Creation is

-- DESCRIPTION :
--   Menu-driven creation of an artificial-parameter polynomial homotopy.

  procedure Default_Homotopy_Settings
                ( k : out positive; a,t : out Complex_Number );
  procedure Default_Homotopy_Settings
                ( k : out positive; a,t : out Complex_Number;
                  proj : out boolean );

  -- DESCRIPTION :
  --   Sets the default settings of the homotopy parameters.

  -- ON RETURN :
  --   k          relaxation constant, by default equals 2;
  --   a          randomly chosen complex number of radius 1;
  --   t          1 is default target value of the continuation parameter;
  --   proj       no projective transformation, proj = false.

  procedure Menu_for_Homotopy_Settings
                ( file : in file_type; k : in out positive;
                  a,t : in out Complex_Number; proj : in out boolean );
  procedure Menu_for_Homotopy_Settings
                ( file : in file_type; k : in out positive;
                  a,t : in out Complex_Number );

  -- DESCRIPTION :
  --   Interactive setting of the homotopy parameters.
  --   If proj is left out, then no projective transformation is set.

  procedure Driver_for_Homotopy_Construction
               ( file : in file_type; p,q : in out Poly_Sys;
                 qsols : in out Solution_List; target : out Complex_Number );

  -- DESCRIPTION :
  --   This is an interactive driver for the construction of an artificial
  --   parameter homotopy.  The user can ask to homogenize the polynomial
  --   system  and can determine the homotopy parameters and the target
  --   value for the  continuation parameter.
  --   This target value is returned as output.
  --   Note that the starting value for the continuation parameter is stored
  --   with the solutions.

end Drivers_for_Homotopy_Creation;
